' Scroll16b-05-opt.bas (PIC16F628A, 4 MHz, PDS 3.5.4.5)
' Copyright (c) by zuisti, all rights reserved!

'Version 0.1, 967 (C: 1046) words, 55.3 (C: 46.8) Hz refresh rate
' pure basic (best portable)
' array[index] access method is used (no pointer)

'version 0.2 not converted to basic because no pointers in PDS

'version 0.3, 955 (C: 1010) words, 70 (C: 60.4) Hz refresh rate (a significant improvement)
' the same as 0.2 in C but indirect array access by FSR and INDF (legal in PDS)
' FSR is a char pointer (hardware char *), and INDF is the *FSR (hw too)
' only one FSR exists in a PIC16 (except the enhanced types) but three in PIC18

'version 0.4, 944 (C: 990) words, 85.6 (C: 83.2) Hz refresh rate (a new significant improvement)
' the same as 0.3 but the exact asm "rlf INDF" is also used (only twice),
' rather than to simulate the rotate with carry (as before). So it's not a pure basic ...

'version 0.5 (opt), 896 words, 120.5 Hz refresh rate (a big significant improvement)
' full hand optimalization with a lot of asm instructions, sorry (zuisti :-)
' the only once called subroutines are placed to their calling place (no gosub - return)

'version 0.6, 884 words, 124 Hz refresh rate
' the same as 0.5 but the pattbuf array is merged into the linebuf array (faster rotating)
' not as useful in this simple case, shorter (-12 byte) and faster a bit but more complex

        Device = 16F628A 
        Xtal = 4
        Optimiser_Level = 3

    Declare FSR_CONTEXT_SAVE  Off

    On_Hardware_Interrupt GoTo Interrupt_Routine

$define SCROLLTIME 6 ' loop count value, the smaller is faster (fastest: 1)

' hw defines:
$define COMMON_CATHODE       'common cathode rows, common anode columns !!!

$define DATA_PIN  PORTB.3 ' the hc595 data

'Store pin: PORTA.3  (RA3), positive clock for shiftregs output storage
$define StorePulse PORTA.3 = 1 : PORTA.3 = 0

'clock pin: PORTB.0  (RB0), positive clock for shift in
$define ClockPulse PORTB.0 = 1 : PORTB.0 = 0

$define SCROLL_PIN PORTB.4   ' a jumper for (re)start (open) / stop (closed) scrolling

'Line decoder inputs: PORTA<2:0>, RA0 = LSB
$define DisableLines PORTA = 0b0111  ' = 7, all 1 (the outputs are active low)

'moved to the end
'Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'moved to the end
'message: LData "This is a very-very long message :-) ",0 ' end sign

' pattbuf array is merged in to the linebuf (much faster rotating)
'Dim pattbuf[7]      As Byte ' holds the actual char's line patterns

Dim linebuf[7*7 + 7] As Byte
' first 7 bytes for the first (uppermost) line (row), the upmost pattern byte,
' then  7 bytes for the 2nd line, the 2nd pattern byte, . and so on, lastly: 7th pattern



Dim ch              As Byte
Dim rowcnt          As Byte
Dim bytecnt         As Byte
Dim bitcnt          As Byte
'Dim pat_fsr         As Byte
'Dim buf_fsr         As Byte
'Dim carry_over      As Byte
Dim width           As Byte
Dim scrollcnt       As Byte

Dim wBitmap_Idx     As Word
Dim wMsg_Idx        As Word

Symbol TMR2IF = PIR1.1   'timer2 IT flag (bank 0)
Symbol TMR2ON = T2CON.2  'timer2 on/off (bank 0)

Symbol Zero_fl  = STATUS.2
Symbol Carry_fl = STATUS.0

' main (start of program):

' HW_Init:
    CMCON = 0b00000111 ' switch off comparators
    DisableLines       ' PORTA = 7, also store pin is zeroed
    PORTB = 0          ' also data and clock pins are zeroed
    TRISA = 0          ' PORTA pins are outputs
    TRISB = 0xF2       ' only PORTB.0,2,3 are outputs
    Clear OPTION_REG.7 ' RBPU, portb pullups on 

' init timer2 for line timing
'    one row display = 630 us (without timer2 waiting) but
'    with scroll (worst case = 494 us) 
'    sum = 630 + 494 = 1124 us (min. for timer2)

'    Prescaler=1:1 PostScaler=1:12 Period = 1.128 ms
'    4 MHz: 1.152 ms/line, 8.062 ms the loop, 124.5 Hz refresh rate

    PR2 = 94            ' Timer2 Match value
    T2CON = 0b01011100  ' enable timer2 with 1:12 postscaler, 1:1 prescaler
    TMR2IF = 0          ' clear IT flag
    PIE1 = 2            ' only TMR2IE = 1: enable Timer2 IT

    INTCON = 0xC0 ' Enable global (GIE) and periferial (PEIE) ITs
' end HW_Init

' Clear_Linebuf:
    FSR = VarPtr linebuf ' first byte
    width = 56 ' lenght of buffer
    $ifdef COMMON_CATHODE
        Movlw 0
    $else
        Movlw 255
    $endif
    While 1=1
        Movwf INDF
        Inc FSR
        Decfsz width ' decrement and skip if zero
    Wend ' this is a 'go to back' instruction
' end Clear_Linebuf

    wMsg_Idx = message ' points the first msg char

    'For prompt scroll:
    scrollcnt = 1 
    ' width = 0 ' not need, it is zero already

While 1=1 ' main forever loop

    If SCROLL_PIN = 1 Then ' scroll sum max 494 (was 538) us
        Dec scrollcnt
        If  scrollcnt = 0 Then
            scrollcnt = SCROLLTIME
            If width = 0 Then
' NextChar:
                While 1=1
                    ' ch = LRead8 message[msg_idx]
                        Call GetCh_ToW
                        Movwf ch
                    '
                    If ch <> 0 Then Break   ' not 0 (not end sign)
                    wMsg_Idx = message      ' else back to begin
                Wend
                Inc wMsg_Idx ' prepare to next char
'  end NextChar, character in ch

                ' ch = ch - 32 : width = 6 : If ch = 0 Then width = 3
                    Movlw 32
                    Subwf ch
                    Movlw 6
                    Btfsc Zero_fl
                        Movlw 3
                    Movwf width                        
                '

' Get_Patterns: ' input: (character - 32) in ch
                ' wBitmap_Idx = ch * 7 + bitmap_address
                    Movf ch, w
                    Movwf wBitmap_Idx
                    Clrf wBitmap_IdxH
                    Bcf Carry_fl
                    Rlf wBitmap_Idx,F
                    Rlf wBitmap_IdxH,F
                    Rlf wBitmap_Idx,F
                    Rlf wBitmap_IdxH,F
                    Rlf wBitmap_Idx,F
                    Rlf wBitmap_IdxH,F  ; *8
                
                    Subwf wBitmap_Idx,F    ; - W = *7 ( w = ch-32 ) 
                    Btfss Carry_fl          ; if borrow, C=0 !!!
                        Decf wBitmap_IdxH,F
                
                    Movlw bitmap ; (bitmap & 255)
                    Addwf wBitmap_Idx,F
                    Movlw High bitmap ; ((bitmap >> 8) & 255)
                    Btfsc Carry_fl
                        Addlw 1
                    Addwf wBitmap_IdxH,F
                '            
                
                ' FSR = VarPtr linebuf + 7 ' points the first pattern byte
                    Movlw linebuf + 7
                    Movwf FSR
                '
                rowcnt = 7 'we have 7 patterns (0...6)
                While 1=1
                    ' read pattern from the bitmap table
                    ' ch = LRead8 bitmap[bitmap_idx] 
                        Call GetTableItem ' using wBitmap_Idx as a pointer
                        Movwf ch
                    '
                    ' drop unused two bits (7,6)
                    ' ch = ch << 2 
                        Rlf ch
                        Rlf ch
                    '
                    ' put it to the pattern byte in linebuf
                    $ifdef COMMON_CATHODE
                        INDF = ~ch ' inverted (0<>1)
                    $else
                        INDF = ch
                    $endif
                    FSR = FSR + 8 ' skip 7 line bytes, points the next patt
                    Inc wBitmap_Idx
                    Decfsz rowcnt
                Wend
'  end Get_Patterns

            EndIf ' width = 0
            Dec width

' Shift_In_A_Column:
            ' FSR = VarPtr linebuf + 55  ' last (7th) pattern
                Movlw linebuf + 55
                Movwf FSR
            '
            bytecnt = 56 ' lenght of buffer, rotate all
            While 1=1 
                Rlf INDF
                Dec FSR
                Decfsz bytecnt
            Wend
' end Shift_In_A_Column

        EndIf ' scrollcnt = 0
    EndIf ' SCROLL_PIN = 1 

' Display: ' loop = 4.403 ms (without timer2 waiting !), 4.403/7 = appr 629 us / row
    FSR = VarPtr linebuf ' points the first byte    

    rowcnt = 0 ' loop until rowcnt = 7
    Repeat ' 629 (was 619) us / row
        bytecnt = 7
        While 1=1
            ch = INDF
            bitcnt = 8
            While 1=1
                ' DATA_PIN = ch.7                
                    Btfss ch.7
                        Clear DATA_PIN       
                    Btfsc ch.7
                        Set DATA_PIN
                '       
                ' ch = ch << 1
                    Rlf ch
                '                    
                ClockPulse
                Decfsz bitcnt
            Wend
            Inc FSR
            Decfsz bytecnt
        Wend
        Inc FSR ' skip pattern byte
        
        'timed display (same row brightness)
        Repeat Until TMR2ON = 0 'wait for timer2 IT (timer2 and all lines off)
        TMR2ON = 1 'restart timer2, also a time gap to switch off drivers
        StorePulse
        PORTA = rowcnt ' switch on the actual line driver
        Inc rowcnt
    Until rowcnt = 7
'  end Display

Wend ' main loop

' subroutines

Interrupt_Routine:
    Context Save
        DisableLines   ' PORTA = 7
        TMR2ON = 0 ' disable timer2
        TMR2IF = 0 ' clear IT flag
  Context Restore

GetCh_ToW:
    Movf wMsg_IdxH, w
    Movwf PCLATH
    Movf wMsg_Idx, w
 Movwf PCL ' go to the proper retlw instr, returns with char in w reg

GetTableItem:
    Movf wBitmap_IdxH,W
    Movwf PCLATH
    Movf wBitmap_Idx,W
 Movwf PCL ' go to the proper retlw instr, returns with pattern byte in w reg

' -------

message: LData "This is a very-very long message :-) ",0 ' end sign

Include "bitmap7.inc" ' ASCII font bitmap table, 7 bytes/char

'------ End
